# -*- coding: utf-8 -*-

from boto3.session import Session, botocore
from storage_provisioner.storage import Storage, S3Storage, AWSS3Region


class StorageProvisioner(object):
    """
        Abstract class. Creates and provisions storage endpoints for arbitrary client data.
    """

    def __init__(self):
        pass

    def provision_storage(self) -> Storage:
        raise NotImplementedError


# region Amazon AWS Constants

DEFAULT_AWS_S3_REGION = AWSS3Region.USWest1

DEFAULT_AWS_S3_POLICY_TEMPLATE = """{
   "Version":"2012-10-17",
   "Statement":[
      {
         "Effect":"Allow",
         "Action":[
            "s3:PutObject",
            "s3:PutObjectAcl",
            "s3:PutObjectAclVersion",
            "s3:GetObject",
            "s3:GetObjectVersion",
            "s3:DeleteObject",
            "s3:DeleteObjectVersion"
         ],
         "Resource":"arn:aws:s3:::{bucket}/{path}*"
      },
      {
         "Effect":"Allow",
         "Action":[
            "s3:ListBucket",
            "s3:GetBucketLocation",
            "s3:ListAllMyBuckets"
         ],
         "Resource":"arn:aws:s3:::{bucket}/{path}"
      }
   ]
}
"""


# endregion

# region Amazon AWS S3 Provisioner


class S3StorageProvisioner:
    """
        Creates and provisions AWS S3 storage endpoints for arbitrary client data.
    """

    def __init__(self,
                 aws_access_key_id: str,
                 aws_secret_access_key: str,
                 default_region: AWSS3Region = DEFAULT_AWS_S3_REGION,
                 default_policy: str = None):
        """
        These are the minimum parameters needed by boto3 in order to provision new S3 buckets and IAM users.

        :param aws_access_key_id:
        :param aws_secret_access_key:
        :param default_region:
        :param default_policy:
        :return:
        """

        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.default_region = default_region
        self.default_policy = default_policy

    def create_federation_token(self,
                                session: Session,
                                user_name: str,
                                user_policy: str,
                                duration_sec: int = 129600,
                                ) -> dict:
        sts = session.client('sts')
        token_resp = sts.get_federation_token(Name=user_name[:32],
                                              Policy=user_policy,
                                              DurationSeconds=duration_sec)
        return token_resp

    def create_bucket_if_needed(self,
                                session: Session,
                                bucket_name: str,
                                region: AWSS3Region,
                                bucket_policy: str = None, ):
        s3 = session.resource('s3')

        bucket_exists = True
        try:
            s3.meta.client.head_bucket(Bucket=bucket_name)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = int(e.response['Error']['Code'])
            if error_code == 404:
                bucket_exists = False

        if not bucket_exists:
            s3.create_bucket(Bucket=bucket_name,
                             CreateBucketConfiguration={'LocationConstraint': region.value})

    def provision_storage(self,
                          user_name: str,
                          bucket_name: str,
                          path: str = None,
                          region: AWSS3Region = None,
                          user_policy: str = None,
                          duration_sec: int = 129600) -> S3Storage:
        """
        This method provisions a new IAM users capable of read/write access to the S3 bucket and path, creating the
        bucket if necessary.

        This method must be called with one of :param policy OR :param path set.
        If policy is None, a default policy with PUT / GET / DELETE access to :param path within :param bucket_name
        will be used.

        :param user_name: the user name to associate with this credential in AWS. Will be truncated to 32 characters.
        :param bucket_name: the S3 bucket name. A bucket with this name will be created if necessary.
        :param path: a path within the bucket, omitting leading '/', where access is scoped if policy is None.
        if a trailing slash is omitted (e.g: 'example'), access is granted to 'example.ext' and 'example/whatever.ext'
        :param region: the region where the S3 bucket should be created if necessary. DEFAULT_AWS_S3_REGION if None.
        :param user_policy: a custom AWS access policy which will be used as-is to scope credentials.
        :param duration_sec: the duration of the returned credentials. If using root AWS credentials, the maximum
        duration is one hour. If an IAM user is used, the maximum duration is 36 hours. If None the maximum available
        duration will be used.
        :return:
        """

        # TODO : Sanitize arguments?

        if region is None:
            region = self.default_region

        region_name = region.value

        session = Session(aws_access_key_id=self.aws_access_key_id,
                          aws_secret_access_key=self.aws_secret_access_key,
                          region_name=region_name)

        if user_policy is None:
            user_policy = DEFAULT_AWS_S3_POLICY_TEMPLATE.replace('{bucket}', bucket_name).replace('{path}', path)

        self.create_bucket_if_needed(bucket_name=bucket_name,
                                     session=session,
                                     region=region)

        token_resp = self.create_federation_token(session=session,
                                                  user_name=user_name,
                                                  user_policy=user_policy,
                                                  duration_sec=duration_sec)

        token_aws_creds = token_resp['Credentials']
        token_aws_federated_user = token_resp['FederatedUser']

        return S3Storage(bucket_name,
                         region_name,
                         path,
                         token_aws_creds['AccessKeyId'],
                         token_aws_creds['SecretAccessKey'],
                         token_aws_creds['SessionToken'],
                         token_aws_creds['Expiration'],
                         token_aws_federated_user['FederatedUserId'],
                         token_aws_federated_user['Arn'],
                         user_policy)

# endregion



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_provisioner
----------------------------------

Tests for `provisioner` module.
"""
import os
import unittest
from boto3.session import Session, botocore
from botocore.exceptions import ClientError
from storage_provisioner.provisioner import S3StorageProvisioner
from storage_provisioner.storage import S3Storage

try:
    from tests.secrets import AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY
except ImportError:
    AWS_ACCESS_KEY_ID = os.getenv('AWS_ACCESS_KEY_ID')
    AWS_SECRET_ACCESS_KEY = os.getenv('AWS_SECRET_ACCESS_KEY')

    if AWS_ACCESS_KEY_ID is None or AWS_SECRET_ACCESS_KEY is None:
        raise EnvironmentError("AWS Credentials not present!")


class TestS3StorageProvisioner(unittest.TestCase):
    s3_provisioner = None
    master_session = None  # Used to perform administrative AWS actions

    test_user_name = 'test_user'
    test_bucket_name = '9a5bb25e-783a-11e5-86ba-b8f6b11601af'
    test_file_name = 'test.txt'
    test_file_contents = 'hello.'.encode()
    tmp_download_file_path = '/tmp/hello.txt'

    def setUp(self):
        print('setup')
        self.s3_provisioner = S3StorageProvisioner(AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)
        self.master_session = Session(aws_access_key_id=AWS_ACCESS_KEY_ID,
                                      aws_secret_access_key=AWS_SECRET_ACCESS_KEY)

    def tearDown(self):
        print('tearDown')
        # Ensure test bucket from previous test is deleted
        self.delete_bucket(self.master_session, self.test_bucket_name)

    def test_provision_storage_default_policy(self):
        """
        Test S3StorageProvisioner's provisioning of storage and credentials with a default user_policy.
        Assert that writing inside the specified storage_path succeeds, and writing outside this path fails.
        """
        storage_path = 'path/to/test/'

        storage = self.s3_provisioner.provision_storage(self.test_user_name,
                                                        self.test_bucket_name,
                                                        storage_path,
                                                        duration_sec=900)

        self.assert_storage_properly_initialized(storage)
        self.assert_storage_policy_valid(storage, storage_path)

    def test_provision_storage_custom_policy(self):

        storage_path = 'test/'

        # string.format requires us to escape braces ('{') by doubling ('{{')
        custom_user_policy = """{{
           "Version":"2012-10-17",
           "Statement":[
              {{
                 "Effect":"Allow",
                 "Action":[
                    "s3:PutObject",
                    "s3:PutObjectAcl",
                    "s3:PutObjectAclVersion",
                    "s3:GetObject",
                    "s3:DeleteObject"
                 ],
                 "Resource":"arn:aws:s3:::{0}/custom/{1}*"
              }},
              {{
                 "Effect":"Allow",
                 "Action":[
                    "s3:ListBucket",
                    "s3:GetBucketLocation",
                    "s3:ListAllMyBuckets"
                 ],
                 "Resource":"arn:aws:s3:::{0}/custom/{1}"
              }}
           ]
        }}
        """.format(self.test_bucket_name, storage_path)

        storage = self.s3_provisioner.provision_storage(self.test_user_name,
                                                        self.test_bucket_name,
                                                        storage_path,
                                                        user_policy=custom_user_policy,
                                                        duration_sec=900)

        # Custom policy restricts access to 'custom' subdirectory
        absolute_storage_path = os.path.join('custom', storage_path)

        self.assert_storage_properly_initialized(storage)
        self.assert_storage_policy_valid(storage, absolute_storage_path)

    def assert_storage_properly_initialized(self, storage: S3Storage):
        """
        Assert that :param storage was properly initialized with test fixture data,
        and that its functions operate as expected.

        :param storage: the Storage object under test
        """

        # Assert that new credentials were issued
        self.assertFalse(storage.aws_access_key_id == AWS_ACCESS_KEY_ID)
        self.assertFalse(storage.aws_secret_access_key == AWS_SECRET_ACCESS_KEY)

        # Federated User Id will be XXX:USERNAME
        self.assertEqual(self.test_user_name, storage.aws_federated_user_id.split(':')[1])

        test_key = 'folder/test_key.txt'
        self.assertEqual(storage.get_url_for_key(test_key),
                         "https://{}.s3.amazonaws.com/{}".format(storage.s3_bucket_name,
                                                                 test_key))

    def assert_storage_policy_valid(self,
                                    storage: S3Storage,
                                    storage_path: str):
        """
        Assert that the credentials in :param storage allow writing only within :param storage_path.

        :param storage: the Storage object providing access credentials
        :param storage_path: the path within which storage's credentials should allow write access
        """

        # Assert that we can write within storage_path
        absolute_storage_path = os.path.join(storage_path, self.test_file_name)

        user_session = Session(storage.aws_access_key_id,
                               storage.aws_secret_access_key,
                               storage.aws_session_token)

        user_s3 = user_session.resource('s3')
        user_s3.Bucket(self.test_bucket_name).put_object(Key=absolute_storage_path, Body=self.test_file_contents)

        user_s3.Bucket(self.test_bucket_name).download_file(absolute_storage_path, self.tmp_download_file_path)

        downloaded_file_contents = open(self.tmp_download_file_path, 'rb').read()

        self.assertEqual(downloaded_file_contents, self.test_file_contents)
        user_s3.Object(self.test_bucket_name, absolute_storage_path).delete()
        os.remove(self.tmp_download_file_path)

        # Assert that we cannot write to the parent directory of storage_path
        try:
            # Ensure the storage_path directory ends with a slash
            path_with_trailing_slash = os.path.join(storage_path, '')
            # The first dirname call removes the trailing slash, the subsequent traverses up a directory
            unscoped_storage_path = os.path.dirname(os.path.dirname(path_with_trailing_slash))
            absolute_unscoped_storage_path = os.path.join(unscoped_storage_path, self.test_file_name)
            user_s3.Bucket(self.test_bucket_name).put_object(Key=absolute_unscoped_storage_path,
                                                             Body=self.test_file_contents)
            raise AssertionError(
                'Wrote to path ({0}) outside storage_path ({1})!'.format(absolute_unscoped_storage_path,
                                                                         storage_path))
        except ClientError as err:
            print('Got ClientError writing outside storage_path, as expected: ', err)

    def delete_bucket(self, session: Session, bucket_name: str):

        s3 = session.resource('s3')

        bucket_exists = True
        try:
            s3.meta.client.head_bucket(Bucket=bucket_name)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = int(e.response['Error']['Code'])
            if error_code == 404:
                bucket_exists = False

        if bucket_exists:
            bucket = s3.Bucket(bucket_name)

            try:
                for key in bucket.objects.all():
                    key.delete()

                bucket.delete()
                print('Deleted bucket {0}'.format(bucket_name))
            except ClientError as err:
                print('Unable to delete bucket:', err)

    if __name__ == '__main__':
        import sys

        sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_storage
----------------------------------

Tests for `storage` module.
"""
import os

import unittest

from storage_provisioner.storage import S3Storage
from storage_provisioner.provisioner import AWSS3Region
try:
    from tests.secrets import AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY
except ImportError:
    AWS_ACCESS_KEY_ID = os.getenv('AWS_ACCESS_KEY_ID')
    AWS_SECRET_ACCESS_KEY = os.getenv('AWS_SECRET_ACCESS_KEY')

    if AWS_ACCESS_KEY_ID is None or AWS_SECRET_ACCESS_KEY is None:
        raise EnvironmentError("AWS Credentials not present!")


class TestS3Storage(unittest.TestCase):

    S3_BUCKET_NAME = 'MrBucket'
    S3_BUCKET_REGION = AWSS3Region.APSouthEast1
    # provided by import:
    # AWS_ACCESS_KEY_ID
    # AWS_SECRET_ACCESS_KEY
    AWS_SESSION_TOKEN = 'TOKE_SESSION'
    AWS_EXPIRATION = 129599
    AWS_FEDERATED_USER_ID = 'FED_USER_ID'
    AWS_ARN = 'ARN'
    AWS_POLICY = 'policy'

    s3_storage = None

    def setUp(self):
        self.s3_storage = S3Storage(self.S3_BUCKET_NAME,
                                    self.S3_BUCKET_REGION,
                                    AWS_ACCESS_KEY_ID,
                                    AWS_SECRET_ACCESS_KEY,
                                    self.AWS_SESSION_TOKEN,
                                    self.AWS_EXPIRATION,
                                    self.AWS_FEDERATED_USER_ID,
                                    self.AWS_ARN,
                                    self.AWS_POLICY)
        pass

    def tearDown(self):
        pass

    def test_s3_storage(self):
        self.assertEqual(self.s3_storage.s3_bucket_name, self.S3_BUCKET_NAME)
        self.assertEqual(self.s3_storage.s3_bucket_region, self.S3_BUCKET_REGION)
        self.assertEqual(self.s3_storage.aws_access_key_id, AWS_ACCESS_KEY_ID)
        self.assertEqual(self.s3_storage.aws_secret_access_key, AWS_SECRET_ACCESS_KEY)
        self.assertEqual(self.s3_storage.aws_session_token, self.AWS_SESSION_TOKEN)
        self.assertEqual(self.s3_storage.aws_expiration, self.AWS_EXPIRATION)
        self.assertEqual(self.s3_storage.aws_federated_user_id, self.AWS_FEDERATED_USER_ID)
        self.assertEqual(self.s3_storage.aws_arn, self.AWS_ARN)

if __name__ == '__main__':
    import sys

    sys.exit(unittest.main())



# -*- coding: utf-8 -*-
from enum import Enum


class Storage(object):
    """
        Represents the credentials needed to access storage.
    """

    def __init__(self):
        pass


class LocalFileStorage(Storage):
    """
        Represents the credentials needed to access storage on the local filesystem.
    """

    def __init__(self, base_path: str=None):
        self.base_path = base_path
        Storage.__init__(self)
        pass


# region Amazon AWS

class AWSS3Region(Enum):
    USEast1 = 'us-east-1'
    USWest1 = 'us-west-1'
    USWest2 = 'us-west-2'

    EUWest1 = 'eu-west-1'
    EUCentral1 = 'eu-central-1'

    APSouthEast1 = 'ap-southeast-1'
    APSouthEast2 = 'ap-southeast-2'
    APNorthEast1 = 'ap-northeast-1'

    SouthAmerica1 = 'sa-east-1'


class AWSFederatedUserMixin(object):
    """
        Represents an AWS Federated User obtained by sts.get_federation_token(..)
    """

    def __init__(self,
                 aws_federated_user_id: str,
                 aws_arn: str,
                 aws_policy: str):

        self.aws_federated_user_id = aws_federated_user_id
        self.aws_arn = aws_arn
        self.aws_policy = aws_policy


class AWSCredentialMixin(object):
    """
        Represents the credentials needed for an AWS FederationToken
    """

    def __init__(self,
                 aws_access_key_id: str,
                 aws_secret_access_key: str,
                 aws_session_token: str,
                 aws_expiration: int):

        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.aws_expiration = aws_expiration


class S3Storage(Storage, AWSCredentialMixin, AWSFederatedUserMixin):
    """
        Represents an AWS FederationToken granting access to an S3 data resource

        {
            'Credentials': {
                'AccessKeyId': 'string',
                'SecretAccessKey': 'string',
                'SessionToken': 'string',
                'Expiration': datetime(2015, 1, 1)
            },
            'FederatedUser': {
                'FederatedUserId': 'string',
                'Arn': 'string'
            },
            'PackedPolicySize': 123
        }
    """

    def __init__(self,
                 s3_bucket_name: str,
                 s3_bucket_region: AWSS3Region,
                 s3_bucket_path: str,
                 aws_access_key_id: str,
                 aws_secret_access_key: str,
                 aws_session_token: str,
                 aws_expiration: int,
                 aws_federated_user_id: str,
                 aws_arn: str,
                 aws_policy: str):

        self.s3_bucket_name = s3_bucket_name
        self.s3_bucket_region = s3_bucket_region
        self.s3_bucket_path = s3_bucket_path

        AWSCredentialMixin.__init__(self,
                                    aws_access_key_id,
                                    aws_secret_access_key,
                                    aws_session_token,
                                    aws_expiration)

        AWSFederatedUserMixin.__init__(self,
                                       aws_federated_user_id,
                                       aws_arn,
                                       aws_policy)

    def get_url_for_key(self, key: str) -> str:
        """
        Return a string representing the URL where the specified key would reside
        :param key: A key describing a file location, leading slash omitted. e.g: "some_folder/myfile.txt"
        :return: a fully-qualified url representing the key. e.g: "https://bucket.s3.amazonaws.com/some_folder/myfile.txt"
        """
        return "https://{}.s3.amazonaws.com/{}".format(self.s3_bucket_name, key)

# endregion



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

with open(os.path.join(os.path.dirname(__file__), 'requirements.txt')) as f:
    requirements = f.read().splitlines()

setup(
    name='storage_provisioner',
    version='0.1.0',
    description="Provisions storage on pluggable backends, such as AWS S3.",
    long_description=readme + '\n\n' + history,
    author="Chris Ballinger, David Brodsky",
    author_email='chrisballinger@gmail.com, dbro@dbro.pro',
    url='https://github.com/PerchLive/storage_provisioner',
    download_url='https://github.com/PerchLive/storage_provisioner/tarball/0.1',
    packages=[
        'storage_provisioner',
    ],
    package_dir={'storage_provisioner':
                 'storage_provisioner'},
    include_package_data=True,
    install_requires=requirements,
    license="Apache License 2.0",
    zip_safe=False,
    keywords='storage_provisioner, aws, boto',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=requirements
)
